
local major, minor = "NickTag-1.0", 1
local NickTag, oldminor = LibStub:NewLibrary (major, minor)

if (not NickTag) then 
	return 
end

_G.NickTag = NickTag

local CONST_INDEX_NICKNAME = 1
local CONST_INDEX_COLOR = 2
local CONST_INDEX_TEXTUREPATH = 3
local CONST_INDEX_TEXTURECOORD = 4
local CONST_INDEX_REVISION = 5

local CONST_COMM_FULLPERSONA = 1
local CONST_COMM_LOGONREVISION = 2
local CONST_COMM_REQUESTPERSONA = 3

NickTag.pool_table = {}
local pool = NickTag.pool_table --> pointer

NickTag.callback_table = {}
NickTag.ini = false

LibStub:GetLibrary ("AceComm-3.0"):Embed (NickTag)
LibStub:GetLibrary ("AceSerializer-3.0"):Embed (NickTag)
LibStub:GetLibrary ("AceTimer-3.0"):Embed (NickTag)

local CallbackHandler = LibStub:GetLibrary ("CallbackHandler-1.0")
NickTag.callbacks = NickTag.callbacks or CallbackHandler:New (NickTag)

function NickTag:OnReceiveComm (_, data, _, source)

	local type, serial, arg3, name, realm =  select (2, NickTag:Deserialize (data))
	
	--> 0x1: received a full persona
	if (type == CONST_COMM_FULLPERSONA) then
		local receivedPersona = arg3
		
		if (serial ~= NickTag:GetSerial()) then

			local storedPersona = NickTag:Get (serial)
			if (not storedPersona) then
				storedPersona = NickTag:Create (serial)
			end
			
			if (storedPersona [CONST_INDEX_REVISION] < receivedPersona [CONST_INDEX_REVISION]) then
				storedPersona [CONST_INDEX_REVISION] = receivedPersona [CONST_INDEX_REVISION]
				storedPersona [CONST_INDEX_NICKNAME] = receivedPersona [CONST_INDEX_NICKNAME]
				storedPersona [CONST_INDEX_COLOR] = receivedPersona [CONST_INDEX_COLOR]
				storedPersona [CONST_INDEX_TEXTUREPATH] = receivedPersona [CONST_INDEX_TEXTUREPATH]
				storedPersona [CONST_INDEX_TEXTURECOORD] = receivedPersona [CONST_INDEX_TEXTURECOORD]
			end
		end
	
	--> 0x2: received a revision version from a guy which logon in the game
	elseif (type == CONST_COMM_LOGONREVISION) then
	
		local receivedRevision = arg3
		local storedPersona = NickTag:Get (serial)
		
		if (not storedPersona or storedPersona [CONST_INDEX_REVISION] < receivedRevision) then
			--> not sure how connected realms will work, but guess this will be fine
			if (realm ~= GetRealmName()) then
				name = name .. "-" .. realm
			end
			NickTag:ScheduleTimer ("RequestPersona", math.random (1, 20), name)
		end
		
	--> 0x3: someone requested my persona, so i need to send to him
	elseif (type == CONST_COMM_REQUESTPERSONA) then
		--> not sure how connected realms will work, but guess this will be fine
		if (realm ~= GetRealmName()) then
			name = name .. "-" .. realm
		end
		NickTag:SendPersona (name)
	end

end

NickTag:RegisterComm ("NickTag", "OnReceiveComm")

function NickTag:SendRevision()
	local battlegroup_serial = NickTag:GetSerial()
	if (not battlegroup_serial) then
		return
	end
	
	local myPersona = NickTag:Get (battlegroup_serial)
	if (myPersona) then
		NickTag:SendCommMessage ("NickTag", NickTag:Serialize (CONST_COMM_LOGONREVISION, battlegroup_serial, myPersona [CONST_INDEX_REVISION], UnitName ("player"), GetRealmName()), "GUILD")
	end
end

function NickTag:OnEvent (event, arg1)
	if (event == "ADDON_LOADED") then
		if (arg1 == "!NickTag") then
			NickTag.pool_table = _G.NickTagCache or {}
			pool = NickTag.pool_table
			NickTag:ScheduleTimer ("SendRevision", 30)
		end
	elseif (event == "PLAYER_LOGOUT") then
		_G.NickTagCache = NickTag.pool_table or {}
	end
end

local frame = CreateFrame ("frame", nil, UIParent)
frame:Hide()
frame:RegisterEvent ("ADDON_LOADED");
frame:RegisterEvent ("PLAYER_LOGOUT");
frame:SetScript ("OnEvent", NickTag.OnEvent)

function NickTag:RequestPersona (target)
	NickTag:SendCommMessage ("NickTag", NickTag:Serialize (CONST_COMM_REQUESTPERSONA, 0, 0, UnitName ("player"), GetRealmName()), "WHISPER", target)
end

function NickTag:SendPersona (target)
	local battlegroup_serial = NickTag:GetSerial()
	if (not battlegroup_serial) then
		return
	end
	
	if (target) then
		NickTag:SendCommMessage ("NickTag", NickTag:Serialize (CONST_COMM_FULLPERSONA, battlegroup_serial, NickTag:Get (battlegroup_serial)), "WHISPER", target)
	else
		NickTag.send_scheduled = false
		NickTag:SendCommMessage ("NickTag", NickTag:Serialize (CONST_COMM_FULLPERSONA, battlegroup_serial, NickTag:Get (battlegroup_serial)), "GUILD")
	end
end

function NickTag:SetNickname (name)
	--> check data before
	assert (type (name) == "string", "NickTag 'SetNickname' expects a string on #1 argument.")
	
	local titlecase = function (first, rest)
		return first:upper()..rest:lower()
	end
	name = name:gsub ("(%a)([%w_']*)", titlecase)
	
	local battlegroup_serial = NickTag:GetSerial()
	if (not battlegroup_serial) then
		return
	end
	
	local nick_table = NickTag:Get (battlegroup_serial)
	if (not nick_table) then
		nick_table = NickTag:Create (battlegroup_serial, true)
	end
	
	nick_table [CONST_INDEX_NICKNAME] = name
	
	self.callbacks:Fire ("NickTag_Update", CONST_INDEX_NICKNAME)
	
	if (not NickTag.send_scheduled) then
		NickTag.send_scheduled = true
		NickTag:ScheduleTimer ("SendPersona", 1)
	end
	
	return true
end

function NickTag:SetColor (r, g, b)
	if (type (r) == "table") then
		r, g, b = unpack (r)
	end
	
	--> check data before
	assert (r and g and b, "NickTag 'SetColor' bad color format. Usage NickTag:SetColor ({r, g, b} or r, g, b)")
	
	local battlegroup_serial = NickTag:GetSerial()
	if (not battlegroup_serial) then
		return
	end
	
	local nick_table = NickTag:Get (battlegroup_serial)
	if (not nick_table) then
		nick_table = NickTag:Create (battlegroup_serial, true)
	end
	
	nick_table [CONST_INDEX_COLOR][1] = r
	nick_table [CONST_INDEX_COLOR][2] = g
	nick_table [CONST_INDEX_COLOR][3] = b
	
	if (not NickTag.send_scheduled) then
		NickTag.send_scheduled = true
		NickTag:ScheduleTimer ("SendPersona", 1)
	end
	
	self.callbacks:Fire ("NickTag_Update", CONST_INDEX_COLOR)
	
	return true
end

function NickTag:SetAvatar (texture, l, r, t, b)

	if (l == nil) then
		l, r, t, b = 0, 1, 0, 1
	elseif (type (l) == "table") then
		l, r, t, b = unpack (l)
	end
	
	--> check data before
	assert (texture and l and r and t and b, "NickTag 'SetAvatar' bad format. Usage NickTag:SetAvatar (texturepath [, L, R, T, B] or texturepath [, {L, R, T, B}])")
	
	local battlegroup_serial = NickTag:GetSerial()
	if (not battlegroup_serial) then
		return
	end
	
	local nick_table = NickTag:Get (battlegroup_serial)
	if (not nick_table) then
		nick_table = NickTag:Create (battlegroup_serial, true)
	end
	
	nick_table [CONST_INDEX_TEXTUREPATH] = texture
	nick_table [CONST_INDEX_TEXTURECOORD][1] = l
	nick_table [CONST_INDEX_TEXTURECOORD][2] = r
	nick_table [CONST_INDEX_TEXTURECOORD][3] = t
	nick_table [CONST_INDEX_TEXTURECOORD][4] = b
	
	if (not NickTag.send_scheduled) then
		NickTag.send_scheduled = true
		NickTag:ScheduleTimer ("SendPersona", 1)
	end
	
	self.callbacks:Fire ("NickTag_Update", CONST_INDEX_TEXTUREPATH)
end

function NickTag:Get (serial)
	--> check data before
	assert (serial, "NickTag 'Get' expects a number on #1 argument.")
	
	if (type (serial) == "string") then
		serial = NickTag:GetSerial (serial)
		if (not serial) then
			return
		end
	end
	
	return pool [serial]
end

function NickTag:GetAvatar (serial, default, silent)
	if (not silent) then
		assert (serial, "NickTag 'GetAvatar' expects a number or string on #1 argument.")
	end
	
	if (type (serial) == "string") then
		serial = NickTag:GetSerial (serial, silent)
		if (serial) then
			local _table = pool [serial]
			if (not _table and default) then
				return default, {0, 1, 0, 1}
			elseif (not _table) then
				return
			end
			return _table [CONST_INDEX_TEXTUREPATH], _table [CONST_INDEX_TEXTURECOORD]
		end
	elseif (type (serial) == "number") then
		local _table = pool [serial]
		if (not _table and default) then
			return default, {0, 1, 0, 1}
		elseif (not _table) then
			return
		end
		return _table [CONST_INDEX_TEXTUREPATH], _table [CONST_INDEX_TEXTURECOORD]
	end
end

function NickTag:GetNickname (serial, default, silent)
	if (not silent) then
		assert (serial, "NickTag 'GetNickname' expects a number or string on #1 argument.")
	end
	
	if (type (serial) == "string") then
		serial = NickTag:GetSerial (serial, silent)
		if (serial) then
			return pool [serial] and pool [serial][CONST_INDEX_NICKNAME] or default or nil
		end
	elseif (type (serial) == "number") then
		return pool [serial] and pool [serial][CONST_INDEX_NICKNAME] or default or nil
	end
end

function NickTag:Create (serial, isSelf)
	--> check data before
	assert (type (serial) == "number", "NickTag 'Create' expects a number on #1 argument.")
	
	--> check if alredy exists
	local alredy_have = pool [serial]
	if (alredy_have) then
		return alredy_have
	end
	
	--> create the table: name, color, texture path, texture texcoord, revision
	local n = { UnitName ("player"), {1, 1, 1}, [[]], {0, 1, 0, 1}, 1 }
	
	--> if not my persona, set revision to 0, this make always get update after creation
	if (not isSelf) then
		n [CONST_INDEX_REVISION] = 0
	end
	
	pool [serial] = n
	return n
end

function NickTag:GetSerial (serial, silent)
	if (not serial) then
		local guid = UnitGUID ("player")
		if (not guid) then
			return
		end
		serial = guid:sub (12, 18)
	else
		if (not silent) then
			assert (type (serial) == "string", "NickTag 'GetSerial' expects a GUID string on #1 parameter"..serial)
			assert (string.len (serial) > 17, "NickTag 'GetSerial' expects a GUID string on #1 parameter")
		else
			if (type (serial) ~= "string") then
				return
			elseif (string.len (serial) < 17) then
				return
			end
		end
		
		serial = serial:sub (12, 18)
	end
	return tonumber ("0x"..serial)
end